# -*- coding: iso-8859-1 -*-

# ==============================================================================
# COPYRIGHT (C) 1991 - 2003  EDF R&D                  WWW.CODE-ASTER.ORG
# THIS PROGRAM IS FREE SOFTWARE; YOU CAN REDISTRIBUTE IT AND/OR MODIFY
# IT UNDER THE TERMS OF THE GNU GENERAL PUBLIC LICENSE AS PUBLISHED BY
# THE FREE SOFTWARE FOUNDATION; EITHER VERSION 2 OF THE LICENSE, OR
# (AT YOUR OPTION) ANY LATER VERSION.
#
# THIS PROGRAM IS DISTRIBUTED IN THE HOPE THAT IT WILL BE USEFUL, BUT
# WITHOUT ANY WARRANTY; WITHOUT EVEN THE IMPLIED WARRANTY OF
# MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE. SEE THE GNU
# GENERAL PUBLIC LICENSE FOR MORE DETAILS.
#
# YOU SHOULD HAVE RECEIVED A COPY OF THE GNU GENERAL PUBLIC LICENSE
# ALONG WITH THIS PROGRAM; IF NOT, WRITE TO EDF R&D CODE_ASTER,
#    1 AVENUE DU GENERAL DE GAULLE, 92141 CLAMART CEDEX, FRANCE.
# ==============================================================================

# $Id: as_system.py,v 1.24 2005/01/19 13:49:13 mcourtoi Exp $
# $Name: v1_2_1 $

# ----- differ messages translation
def _(mesg): return mesg

import string
import re
import os
import sys
import os.path
import popen2
import shutil
import glob
from types import StringType,DictType

#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------
class ASTER_SYSTEM:
   """Class to encapsultate "system" commands on local or remote machine.
   Available methods :
      'Copy','Shell','IsRemote','MkDir','Delete','Exists', 'IsDir',
      'Gzip', 'Gunzip', 'AddToEnv',
      'VerbStart, 'VerbEnd', 'VerbIgnore'
   Supported protocols :
   - for copy :
      LOCAL, RCP, SCP, RSYNC, HTTP
   - for shell execution :
      LOCAL, RSH, SSH
   """
   # this value must be set during installation step.
   MaxCmdLen=1024
   # line length -9
   _LineLen=80-9
   # required parameters for initialisation
   RequiredParam=['remote_copy_protocol','remote_shell_protocol']
#-------------------------------------------------------------------------------
   def __init__(self, run, **kargs):
      """run : ASTER_RUN object
      OR just a dictionnary to define 'RequiredParam' + 'verbose', 'debug'
      """
      self.verbose=run['verbose']
      self.debug=run['debug']
      if kargs.has_key('maxcmdlen'):
         self.MaxCmdLen = kargs['maxcmdlen']
      # fill param
      self.param={}
      for key in self.RequiredParam:
         self.param[key]=run[key]
      # ----- reference to ASTER_RUN object which manages the execution
      self.run=run
      # informations
      self.info= {
         # store local user (will be the default for remote machines)
         'local_user' : os.environ['LOGNAME'],
         'local_host' : os.uname()[1],
      }
      # check if protocols are supported
      if not self.param['remote_shell_protocol'] in ('LOCAL','RSH','SSH'):
         self._mess(_('unknown protocol : %s') % self.param['remote_shell_protocol'],
               _('<A>_ALARM'))
      if not self.param['remote_copy_protocol'] in ('LOCAL','RCP','SCP','RSYNC','HTTP'):
         self._mess(_('unknown protocol : %s') % self.param['remote_copy_protocol'],
               _('<A>_ALARM'))
      if self.debug: print '<DBG> ASTER_SYSTEM param :\n',self.param

#-------------------------------------------------------------------------------
   def _mess(self,msg,cod=''):
      """Just print a message
      """
      if hasattr(self.run,'Mess'):
         self.run.Mess(msg,cod)
      else:
         print '%-18s %s' % (cod,msg)

#-------------------------------------------------------------------------------
   def VerbStart(self,cmd,verbose=None):
      """Start message in verbose mode
      """
      Lm=self._LineLen
      if verbose==None:
         verbose=self.verbose
      if verbose:
         pcmd=cmd
         if len(cmd)>Lm-2 or cmd.count('\n')>0:
            pcmd=pcmd+'\n'+' '*Lm
         print ('%-'+str(Lm)+'s') % (pcmd,),
         sys.stdout.flush()

#-------------------------------------------------------------------------------
   def VerbEnd(self,iret,output='',verbose=None):
      """Ends message in verbose mode
      """
      if verbose==None:
         verbose=self.verbose
      if verbose:
         if iret==0:
            print '[  OK  ]'
         else:
            print _('[FAILED]')
            print _('Exit code : %d') % iret
         if (iret<>0 or self.debug) and output:
            print output

#-------------------------------------------------------------------------------
   def VerbIgnore(self,verbose=None):
      """Ends message in verbose mode
      """
      if verbose==None:
         verbose=self.verbose
      if verbose:
         print '[ SKIP ]'

#-------------------------------------------------------------------------------
   def local_shell(self,cmd,bg=False,verbose=False,follow_output=False,
         alt_comment=None,interact=False):
      """Execute a command shell
         cmd      : command
         bg       : put command in background if True
         verbose  : print status messages during execution if True
         follow_output : follow interactively output of command
         alt_comment : print this "alternative comment" instead of "cmd"
         interact : allow the user to interact with the process
            (don't close stdin). bg=True implies interact=False.
      Return :
         iret     : exit code if bg = False,
                    process id if bg = True
         output   : output lines (as string)
      """
      if not alt_comment:
         alt_comment=cmd
      if bg:
         interact=False
      if len(cmd)>self.MaxCmdLen:
         self._mess((_('length of command shell greater '\
               'than %d characters.') % self.MaxCmdLen),_('<A>_ALARM'))
      if self.debug:
         print '<DBG> <local_shell>',cmd
         print '<DBG> <local_shell> background mode : ',bg
      # exec
      self.VerbStart(alt_comment,verbose=verbose)
      if follow_output and verbose:
         print _('\nCommand output :')
#MC A REVOIR : il faudrait connecter tochild  stdin et fromchild  stdout
# faut-il le faire dans un thread spar....? plus tard ;-)
      # run interactive command
      if interact:
         iret=os.system(cmd)
         return iret,''
      # use popen to manipulate stdout/stderr
      output=[]
      p=popen2.Popen4(cmd)
      p.tochild.close()
      if not bg:
         if not follow_output:
            output=p.fromchild.readlines()
         else:
            while p.poll()==-1:
               output.append(p.fromchild.readline())
               # \n already here...
               print output[-1],
               sys.stdout.flush()
            # to be sure to empty the buffer
            end=p.fromchild.readlines()
            print ''.join(end)
            output.extend(end)
         iret=p.wait()
      else:
         iret=0
      p.fromchild.close()
      output=''.join(output)

      # repeat header message
      if follow_output:
         self.VerbStart(alt_comment,verbose=verbose)
      mat=re.search('EXIT_CODE=([0-9]+)',output)
      if mat:
         try:
            iret=int(mat.group(1))
         except:
            pass
      self.VerbEnd(iret,output,verbose=verbose)
      if bg:
         iret=p.pid
         if verbose:
            print 'Process ID : ',iret
      return iret,output

#-------------------------------------------------------------------------------
   def filename2dict(self,s):
      """Convert file name 's' given in format [[user[:passwd]@]machine:]name
      into a dictionnary with keys = user, passwd, mach, name
      """
      d={
         'user'   : '',
         'passwd' : '',
         'mach'   : '',
         'name'   : ''
      }
      if not type(s) is StringType:
         self._mess(_('unexpected type'),_('<F>_PROGRAM_ERROR'))
      mat=re.search('(.*):(.*)@(.*):(.*)',s)
      n=4
      if mat==None:
         mat=re.search('(.*)@(.*):(.*)',s)
         n=3
         if mat==None:
            mat=re.search('(.*):(.*)',s)
            n=2
            if mat==None:
               d['name']=s
               n=1
      if n>=2:
         g=list(mat.groups())
         d['name']=g.pop()
         d['mach']=g.pop()
      if n>=3:
         d['user']=g.pop()
      if n>=4:
         d['passwd']=d['user']
         d['user']=g.pop()
      if d['mach']<>'' and d['user']=='':
         d['user']=self.info['local_user']
      #if self.debug: print '<DBG> <filename2dict>',s,d
      return d

#-------------------------------------------------------------------------------
   def Copy(self,dest,*src,**opts):
      """Copy files/directories (from 'src' list) to 'dest'
      Remote names format :
         [[user[:passwd]@]machine:]name
      where name is an absolute or a relative path.
      optionnal arguments : verbose, follow_output passed to local_shell,
         niverr : default is set to <F>_COPY_ERROR, but may be '<A>' or less !
         protocol : to use a different protocol just for this copy.
      Note : if src[i] is a directory and dest doesn't exist
             dest is created and the content of src[i] is copied into dest.
             This is not a default behavior of rsync which would create
             `basename src[i]` in dest even if dest doesn't exist.
      """
      iret=0
      kargs={}
      # take values from opts if exist
      kargs['verbose']=opts.get('verbose',self.verbose)
      kargs['follow_output']=opts.get('follow_output',False)
      if opts.has_key('niverr'):
         niverr=opts['niverr']
      else:
         niverr=_('<F>_COPY_ERROR')
      ddest=self.filename2dict(dest)
      if self.IsRemote(dest):
         fdest=ddest['user']+'@'+ddest['mach']+':'+ddest['name']
      else:
         fdest=ddest['name']
      if self.debug:
         print '<DBG> source list :',src
         print '      destination :',fdest

      if len(src) < 1:
         self._mess(_('no source file to copy'),_('<A>_ALARM'))

      for f in src:
         # here because we can change 'proto' if necessary
         proto=opts.get('protocol',self.param['remote_copy_protocol'])
         jret=0
         df=self.filename2dict(f)
         if self.IsRemote(f):
            fsrc=df['user']+'@'+df['mach']+':'+df['name']
         else:
            fsrc=df['name']
         cmd=''
         tail='.../'+'/'.join(f.split('/')[-2:])
         kargs['alt_comment']='copying %s' % tail
         if df['mach']=='' and ddest['mach']=='':
            cmd='cp -r '+fsrc+' '+fdest
         else:
            if proto=='RSYNC' and df['mach']<>'' and ddest['mach']<>'':
               proto='RCP'
               self._mess(_("copying a remote file to another remote server " \
                     "isn't allowed through RSYNC, I try using RCP."))
            if proto=='RCP':
               cmd='rcp -r '+fsrc+' '+fdest
            elif proto=='SCP':
               cmd='scp -rBCq '+fsrc+' '+fdest
            elif proto=='RSYNC':
               if self.IsDir(f) and not self.Exists(dest):
                  self.MkDir(dest)
                  cmd ='rsync -rz '+os.path.join(fsrc,'*')+' '+fdest
               else:
                  cmd='rsync -rz '+fsrc+' '+fdest
            elif proto=='HTTP':
               str_user=''
               if not df['user'] in ('','anonymous'):
                  str_user=df['user']+'@'
               # dest must be local
               if ddest['mach']=='':
                  cmd='wget http://'+str_user+df['mach']+df['name']+' -O '+fdest
               else:
                  cmd=''
                  self._mess(_('remote destination not allowed through %s' \
                        ' : %s') % (proto,fdest), niverr)
         if cmd<>'':
            jret,out=self.local_shell(cmd, **kargs)
            if jret<>0 and niverr<>'SILENT':
               self._mess(_('error during copying %s to %s') % (f,fdest), niverr)
         else:
            self._mess(_('unexpected error or unknown protocol : %s') \
                   % proto, niverr)
         iret=max(jret,iret)
      return iret

#-------------------------------------------------------------------------------
   def Shell(self,cmd,mach='',user='',**opts):
      """Execute a command shell on local or remote machine.
      Options : see local_shell
      """
      iret=1
      # valeurs par dfaut
      kargs={
         'verbose'         : self.verbose,
         'bg'              : False,
         'interact'        : False,
         'follow_output'   : False,
         'alt_comment'     : cmd,
      }
      # surcharge par opts
      kargs.update(opts)

      proto=self.param['remote_shell_protocol']
      #if cmd.find('"')<>-1:
      #      self._mess(_("""an error may occurs because of '"' character ' \
      #            'in command"""),_('<A>_ALARM'))
      #cmd="sh -c '"+cmd+"'"
      if mach=='':
         iret,out=self.local_shell(cmd, **kargs)
      else:
         if user=='':
            user=os.environ['LOGNAME']
         action=''
         if proto=='RSH':
            action='rsh -n -l '+user+' '+mach+' '+cmd
         elif proto=='SSH':
            action='ssh -n -l '+user+' '+mach+' '+cmd
         if action<>'':
            iret,out=self.local_shell(action, **kargs)
         else:
            self._mess(_('unknown protocol : %s') % proto,_('<A>_ALARM'))
      return iret,out

#-------------------------------------------------------------------------------
   def Gzip(self,src,**opts):
      """Compress file or content of directory 'src'.
      optionnal arguments : verbose, niverr, cmd (to override gzip command).
      Only local files/directories supported !
      """
      return self._gzip_gunzip('gzip',src,**opts)

#-------------------------------------------------------------------------------
   def Gunzip(self,src,**opts):
      """Decompress file or content of directory 'src'.
      optionnal arguments : verbose, niverr, cmd (to override gzip -d command).
      Only local files/directories supported !
      """
      return self._gzip_gunzip('gunzip',src,**opts)

#-------------------------------------------------------------------------------
   def _gzip_gunzip(self,mode,src,**opts):
      """Called by Gzip (mode=gzip) and Gunzip (mode=gunzip) methods.
      Return status and filename of src after de/compression.
      Only local files/directories supported !
      """
      para={
         'gzip' : {
            'cmd'     : 'gzip ',
            'niverr'  : _('<F>_COMPRES_ERROR'),
            'msg0'    : _('no file/directory to compress'),
            'comment' : _('compressing %s'),
            'msgerr'  : _('error during compressing %s'),
         },
         'gunzip' : {
            # sometimes gunzip doesn't exist, gzip seems safer
            'cmd'     : 'gzip -d ',
            'niverr'  : _('<F>_DECOMPRESSION'),
            'msg0'    : _('no file/directory to decompress'),
            'comment' : _('decompressing %s'),
            'msgerr'  : _('error during decompressing %s'),
         },
      }
      if not mode in para.keys():
         self._mess(_('unknown mode : %s') % mode,_('<F>_PROGRAM_ERROR'))
      iret=0
      if opts.has_key('cmd'):
         cmd=opts['cmd']
      else:
         cmd=para[mode]['cmd']
      if opts.has_key('verbose'):
         verbose=opts['verbose']
      else:
         verbose=self.verbose
      if opts.has_key('niverr'):
         niverr=opts['niverr']
      else:
         niverr=para[mode]['niverr']

      if len(src) < 1:
         self._mess(para[mode]['msg0'],_('<A>_ALARM'))
         return iret

      if not os.path.isdir(src):
         if mode=='gzip':
            name=src+'.gz'
         else:
            name=re.sub('\.gz$','',src)
         lf=[src]
      else:
         name=src
         lf=glob.glob(os.path.join(src,'*'))
      for f in lf:
         jret=0
         comment=para[mode]['comment'] % f
         if os.path.isdir(f):
            self.VerbStart(comment, verbose=verbose)
            self.VerbIgnore(verbose=verbose)
         else:
            jret,out=self.local_shell(cmd+f,verbose=verbose,alt_comment=comment)
         if jret<>0:
            self._mess(para[mode]['msgerr'] % f, niverr)
         iret=max(iret,jret)
      return iret, name

#-------------------------------------------------------------------------------
   def IsRemote(self,path):
      """Return True if 'path' seems to be on a remote host.
      NB : we suppose that host and host.domain are identical.
      """
      fhost=self.filename2dict(path)['mach'].split('.')[0]
      host=self.info['local_host']
      return ( fhost<>'' and host<>fhost )

#-------------------------------------------------------------------------------
   def Exists(self,path):
      """Return True if 'path' exists (file or directory).
      """
      iret=0
      exists=True
      dico=self.filename2dict(path)
      if self.IsRemote(path):
         cmd="\"sh -c 'if test -f "+dico['name']+' ; then echo FILE_EXISTS; ' \
                         "elif test -d "+dico['name']+' ; then echo DIR_EXISTS ; ' \
                                                         "else echo FALSE ; fi'\""
         iret,out=self.Shell(cmd,dico['mach'],dico['user'],verbose=self.verbose)
         if out.find('FALSE') > -1:   exists=False
      else:
         exists=os.path.exists(dico['name'])
      return exists

#-------------------------------------------------------------------------------
   def IsDir(self,path):
      """Return True if 'path' is a directory.
      """
      iret=0
      isdir=False
      dico=self.filename2dict(path)
      if self.IsRemote(path):
         cmd="\"sh -c 'if test -d "+dico['name']+' ; then echo IS_DIRECTORY; ' \
                                             "else echo FALSE ; fi'\""
         iret,out=self.Shell(cmd,dico['mach'],dico['user'],verbose=self.verbose)
         if out.find('IS_DIRECTORY') > -1:  isdir=True
      else:
         isdir=os.path.isdir(dico['name'])
      return isdir

#-------------------------------------------------------------------------------
   def MkDir(self,rep,niverr=None,verbose=None):
      """Create the directory 'rep' (mkdir -p ...)
      """
      if niverr==None:
         niverr=_('<F>_MKDIR_ERROR')
      if verbose==None:
         verbose=self.verbose
      iret=0
      dico=self.filename2dict(rep)
      if self.IsRemote(rep):
         cmd='mkdir -p '+dico['name']
         iret,out=self.Shell(cmd,dico['mach'],dico['user'],verbose=verbose)
      else:
         self.VerbStart(_('creating directory %s') % dico['name'],verbose)
         # ----- it's not tested in remote
         if self.IsDir(rep):
            iret=0
            self.VerbIgnore(verbose)
         else:
            s=''
            try:
               os.makedirs(dico['name'])
            except OSError, s:
               iret=4
            self.VerbEnd(iret,s,verbose)
      if iret<>0:
         self._mess(_('can not create directory %s') % rep,niverr)
      return iret

#-------------------------------------------------------------------------------
   def Delete(self,rep,verbose=None):
      """Delete an entire directory tree (rm -rf ...)
      """
      if verbose==None:
         verbose=self.verbose
      iret=0
      dico=self.filename2dict(rep)
      # preventing to delete first level directories (as /, /home, /usr...)
      if dico['name'][0]=='/' and len(dico['name'][:-1].split('/'))<=2:
         self._mess(_('deleting this directory seems to dangerous. ' \
               '%s has not been removed') % dico['name'],_('<A>_ALARM'))
      if self.IsRemote(rep):
         cmd='\rm -rf '+dico['name']
         iret,out=self.Shell(cmd,dico['mach'],dico['user'],verbose=self.verbose)
      else:
         self.VerbStart(_('deleting %s') % rep,verbose)
         s=''
         try:
            if os.path.isdir(dico['name']):
               shutil.rmtree(dico['name'])
            elif os.path.exists(dico['name']):
               os.remove(dico['name'])
            else:
               self.VerbIgnore(verbose)
               return iret
         except OSError:
            iret=4
         self.VerbEnd(iret,s,verbose)
      return iret

#-------------------------------------------------------------------------------
   def AddToEnv(self, profile):
      """Read 'profile' file (with sh/bash/ksh syntax) and add updated
      variables to os.environ.
      """
      def env2dict(s):
         """Convert output to a dictionnary."""
         l=s.split(os.linesep)
         d={}
         for line in l:
            mat=re.search('^([-a-zA-Z_0-9@\+]+)=(.*$)', line)
            if mat<>None:
               d[mat.group(1)]=mat.group(2)
         return d
      if not os.path.isfile(profile):
         self._mess(_('file not found : %s') % profile, '<A>_FILE_NOT_FOUND')
         return
      # read initial environment
      iret, out = self.Shell('sh -c env')
      env_init = env2dict(out)
      if iret<>0:
         self._mess(_('error getting environment'), '<E>_ABNORMAL_ABORT')
         return
      # read profile and dump modified environment
      iret, out = self.Shell('sh -c "source '+profile+' ; env"')
      env_prof = env2dict(out)
      if iret<>0:
         self._mess(_('error reading profile : %s') % profile,
               '<E>_ABNORMAL_ABORT')
         return
      # "diff"
      for k, v in env_prof.items():
         if env_init.get(k, None)<>v:
            if self.debug:
               print 'AddToEnv adding : %s=%s' % (k,v)
            os.environ[k]=v

#-------------------------------------------------------------------------------
if __name__ == '__main__':
   tCopy     = False
   tShell    = False
   tIsRemote = True
   tMkDir    = False
   tDelete   = False

   # remote connections parameters
   remote_params={
      'remote_copy_protocol'   : 'RCP',
      'remote_shell_protocol'  : 'RSH',
      'verbose'                : True,
      'debug'                  : True,
   }
   cmd=ASTER_SYSTEM(remote_params)
   #cmd.debug=1

   ltest=['/tmp/interactif.3300',
#          'aster.cla.edfgdf.fr:/home/courtois',
#          'mcourtoi@aster.cla.edfgdf.fr:/aster/eda/regles_EDA.txt',
#          'user:mypass@ftpserver:/pub/incoming',
#          '@aster:/tmp',
          '/peutprobable/que/ce/chemin/existe',
          '130.98.5.6:REST/dev_7.4/TEST']
   dest='/tmp/DEST'
   # test Copy
   if tCopy:
      for s in ltest:
         dd=cmd.filename2dict(s)
         iret=cmd.Copy(dest,s)
      iret=cmd.Copy(dest,*ltest)
   # test Shell
   if tShell:
      c0='ls -l'
      print '\n----- TEST cmd :',c0
      iret,output=cmd.Shell(c0)
      print 'exit code =',iret
      c0='uname -ak'
      print '\n----- TEST verbose=False, cmd :',c0
      iret,output=cmd.Shell(c0,verbose=False)
      print 'exit code =',iret
#       c0='ls -al /root/forbidden ; echo EXIT_CODE=8'
#       print '\n----- TEST remote, cmd :',c0
#       iret,output=cmd.Shell(c0,mach='cli70cx')
#       print 'exit code =',iret
      c0='ls -al /root/forbidden'
      print '\n----- TEST remote, cmd :',c0
      iret,output=cmd.Shell(c0)
      print 'exit code =',iret
      
#       bdir='/usr/local/aster/STA7.4/tcl/'
#       c0=bdir+'as_run '+bdir+'make_etude'
#       print '\n----- TEST command :',c0
#       iret,output=cmd.Shell(c0,verbose=False,follow_output=True)
#       print 'exit code =',iret
   # test IsRemote
   if tIsRemote:
      for f in ltest:
         if cmd.Exists(f): ex='exists'
         else:             ex='not found'
         if cmd.IsRemote(f): loc='remote'
         else:               loc='local'
         if cmd.IsDir(f): di='directory'
         else:            di='not dir'
         print '%-60s : %10s / %10s / %10s' % (f,loc,ex,di)
   # test MkDir
   if tMkDir:
      rep='/root/mc/z'
      print '\n----- MkDir should failed :',rep
      iret=cmd.MkDir(rep)
      print 'iret =',iret

      rep='130.98.5.6:/root/mc/z'
      print '\n----- MkDir should failed (but rsh returns 0) :',rep
      iret=cmd.MkDir(rep)
      print 'iret =',iret

      rep='/tmp/mc/test'
      print '\n----- MkDir',rep
      iret=cmd.MkDir(rep)
      print 'iret =',iret

      cmd.debug=True
      rep='aster@cli70cx:/tmp/test_mkdir/zzz'
      print '\n----- MkDir (remote) :',rep
      iret=cmd.MkDir(rep)
      print 'iret =',iret
      if cmd.IsDir(rep):
         print 'ok'
      else:
         print 'failed'

   # test Delete
   if tDelete:
      rep='/tmp/mc/test'
      print 'Delete',rep
      iret=cmd.Delete(rep)
      print 'iret =',iret
      
      rep='/tmp/test_mkdir'
      print 'Delete',rep
      iret=cmd.Delete(rep)
      print 'iret =',iret
